# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to set baseline PHPP lighting."""

from typing import Optional

from PHX.PHPP.phpp_app import PHPPConnection

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.lighting_space_types import space_type_map


def find_lighting_installed_power(
    _baseline_code: BaselineCode, _program_name: str
) -> Optional[float]:
    """Find the baseline lighting installed power density for a given PHPP space.

    Arguments:
    ----------
        * _baseline_code: BaselineCode
            The BaselineCode object
        * _program_name: str
            The PHPP space type name

    Returns:
    --------
        * code_LPD: Optional[float]
            The baseline lighting installed power density for the given PHPP space type
    """

    # -- First, find the building-code name which corresponds to the PHPP space type
    code_name = space_type_map.get(_program_name.upper().strip(), None)

    if (
        not code_name
        or code_name not in _baseline_code.tables.lighting_area_method.LPD.keys()
    ):
        msg = (
            f"> Warning: The PHPP Space-Type: '{_program_name}' does not have a corresponding "
            "building-code lighting-space-type? Cannot set electric lighting baseline. "
            "You should set the electric lighting installer power density manually on the "
            "PHPP 'Electricity non-res' worksheet."
        )
        print(msg)
        return None

    # -- Then, find the lighting installed power density for that building-code name
    return _baseline_code.tables.lighting_area_method.LPD[code_name]


def set_baseline_lighting_installed_power_density(
    _phpp_conn: PHPPConnection, _baseline_code: BaselineCode
) -> None:
    """Set the PHPP 'Electricity non-res' worksheet baseline lighting installed power density.

    Arguments:
    ----------
        _phpp_conn: phpp_app.PHPPConnection
            The PHPPConnection object
        _baseline_code: BaselineCode
            The BaselineCode object
    """

    for row_num in _phpp_conn.elec_non_res.lighting.used_lighting_row_numbers:
        row_data = _phpp_conn.elec_non_res.lighting.get_lighting_row_data(row_num)
        code_LPD = find_lighting_installed_power(
            _baseline_code, row_data.utilization_profile_name
        )
        if not code_LPD:
            continue
        _phpp_conn.elec_non_res.lighting.set_lighting_power_density(
            row_num, code_LPD, "W/M2"
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to set baseline PHX Model windows."""

from typing import Callable, Dict

from PHX.model.project import PhxProject
from PHX.model.components import PhxComponentAperture
from PHX.model.constructions import PhxConstructionWindow

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.options import ClimateZones, PF_Groups, Use_Groups


def get_baseline_window_u_value(
    _baseline_code: BaselineCode, _climate_zone: ClimateZones, _use_group: Use_Groups
) -> float:
    """Get the baseline U-Value for the specified climate zone."""
    all_u_values = _baseline_code.get_window_u_values()
    cz_uvalues = all_u_values.get_u_value_for_climate(_climate_zone)
    u_value = cz_uvalues.get_u_values_for_use_group(_use_group)
    return u_value


def get_baseline_window_SHGC(
    _baseline_code: BaselineCode, _climate_zone: ClimateZones, _pf_group: PF_Groups
) -> float:
    """Get the baseline SHGC for the specified climate zone and Projection Factor group."""
    all_shgcs = _baseline_code.get_window_shgcs()
    cz_shgcs = all_shgcs.get_shgcs_for_climate(_climate_zone)
    pf_shgc = cz_shgcs.get_shgc_for_pf(_pf_group)
    return pf_shgc


def set_baseline_window_construction(
    _phx_project: PhxProject,
    _baseline_code: BaselineCode,
    _climate_zone: ClimateZones,
    _pf_group: PF_Groups,
    _use_group: Use_Groups,
    _output: Callable = print,
) -> PhxProject:
    # -- Get the baseline values for U-Value and SHGC
    u_value = get_baseline_window_u_value(_baseline_code, _climate_zone, _use_group)
    _output(f"Using baseline window U-Value: {u_value :.3f} for all windows.")

    shgc = get_baseline_window_SHGC(_baseline_code, _climate_zone, _pf_group)
    _output(f"Using baseline window SHGC: {shgc :.2f} for all windows.")
    

    # -----------------------------------------------------------------------------------
    # -- Collect all the unique 'PhxComponentAperture' types in the PHX project
    _output(f"Collecting the unique '{PhxComponentAperture.__name__}' types in the PHX-Project: '{_phx_project.name}'.")
    unique_aperture_types: Dict[str, PhxComponentAperture] = {}
    for variant in _phx_project.variants:
        for ap in variant.building.aperture_components:
            unique_aperture_types[ap.unique_key] = ap


    # -----------------------------------------------------------------------------------
    # -- Create a new baseline 'PhxConstructionWindow' for each of the unique 'PhxComponentAperture' types
    _output(f"Creating a new baseline '{PhxConstructionWindow.__name__}' for each unique '{PhxComponentAperture.__name__}' type.")
    for i, unique_window_type_key in enumerate(unique_aperture_types.keys(), start=1):
        window_type = unique_aperture_types[unique_window_type_key]
        baseline_phx_window = PhxConstructionWindow.from_total_u_value(
            u_value, shgc, f"BASELINE: WINDOW {i :03}"
        )
        baseline_phx_window.id_num_shade = window_type.shade_type_id_num

        # -- Add the new baseline window construction to the PHX project
        _phx_project.add_new_window_type(
            baseline_phx_window, _key=unique_window_type_key
        )


    # -----------------------------------------------------------------------------------
    # -- Set the baseline PhxComponentAperture's 'PhxConstructionWindow'
    _output(f"Setting all '{PhxConstructionWindow.__name__}' to the baseline types.")
    for variant in _phx_project.variants:
        for ap in variant.building.aperture_components:
            baseline_phx_window_type = _phx_project.get_window_type(
                ap.unique_key
            )
            ap.set_window_type(baseline_phx_window_type)

    return _phx_project


def set_baseline_window_area(
    _phx_project: PhxProject, _baseline_code: BaselineCode, _output: Callable = print
) -> PhxProject:
    """Set the maximum window-to-wall ratio in the PHPP Windows Worksheet.

    Arguments:
    ----------
        phpp_conn: phpp_app.PHPPConnection
            The PHPPConnection object
        baseline_code: BaselineCode
            The BaselineCode object
    """

    # -- As per NYS Code, C402.4.1 Maximum Area:
    #
    # > "The vertical fenestration area, not including opaque doors and
    # > opaque spandrel panels, shall be not greater than 30
    # > percent of the GROSS above-grade wall area."
    #
    # -- So be sure to use the GROSS wall area, BEFORE the windows
    # -- have been punched out of them.

    maximum_wwr = _baseline_code.get_baseline_max_wwr()
    current_gross_wall_area = _phx_project.get_total_gross_wall_area()
    maximum_aperture_area = current_gross_wall_area * maximum_wwr
    current_aperture_area = _phx_project.get_total_wall_aperture_area()
    current_wwr = current_aperture_area / current_gross_wall_area

    _output(f"Current wall aperture area = {current_aperture_area :.1f}")
    _output(f"Current gross wall area = {current_gross_wall_area:.1f}")
    _output(
        f"Wall aperture area limit = {maximum_aperture_area:.1f} [{current_gross_wall_area :.1f} x {maximum_wwr :.0%}]"
    )

    if current_wwr < maximum_wwr:
        _output(
            f"Current WWR {current_wwr:.2%} is less than baseline maximum of {maximum_wwr:.0%}."
        )
        return _phx_project

    aperture_area_reduction_required = current_aperture_area - maximum_aperture_area
    _output(
        f"Wall aperture area reduction required = {aperture_area_reduction_required:.1f}"
    )

    # -- Figure out the right scale factor to use
    area_scale_factor = 1 - (aperture_area_reduction_required / current_aperture_area)
    geometric_scale_factor = (area_scale_factor) ** 0.5
    _output(
        f"Scaling wall apertures down to {area_scale_factor :.1%} of original size."
    )

    # -- Scale all the Aperture components
    for variant in _phx_project.variants:
        variant.building.scale_all_wall_aperture_components(geometric_scale_factor)

    return _phx_project


def set_baseline_skylight_area(
    _phx_project: PhxProject, _baseline_code: BaselineCode, output: Callable = print
) -> PhxProject:
    maximum_srr = _baseline_code.get_baseline_max_srr()
    current_gross_roof_area = _phx_project.get_total_gross_roof_area()
    maximum_aperture_area = current_gross_roof_area * maximum_srr
    current_aperture_area = _phx_project.get_total_roof_aperture_area()
    current_srr = current_aperture_area / current_gross_roof_area

    output(f"Current roof aperture area = {current_aperture_area :.1f}")
    output(f"Current gross roof area = {current_gross_roof_area:.1f}")
    output(
        f"Roof aperture area limit = {maximum_aperture_area:.1f} [{current_gross_roof_area :.1f} x {maximum_srr :.0%}]"
    )

    if current_srr < maximum_srr:
        output(
            f"Current SRR {current_srr:.2%} is less than baseline maximum of {maximum_srr:.0%}."
        )
        return _phx_project

    aperture_area_reduction_required = current_aperture_area - maximum_aperture_area
    output(
        f"Roof aperture area reduction required = {aperture_area_reduction_required :.1f}"
    )

    # -- Figure out the right scale factor to use
    area_scale_factor = 1 - (aperture_area_reduction_required / current_aperture_area)
    geometric_scale_factor = (area_scale_factor) ** 0.5
    output(f"Scaling roof apertures down to {area_scale_factor :.1%} of original size.")

    # -- Scale all the Aperture components
    for variant in _phx_project.variants:
        variant.building.scale_all_roof_aperture_components(geometric_scale_factor)

    return _phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for creating and setting Baseline Constructions of a PHX model."""

from copy import copy
from typing import NamedTuple, Callable

from PHX.model.constructions import PhxConstructionOpaque

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.options import ClimateZones, Use_Groups


class BaselinePHXConstructions(NamedTuple):
    """NamedTuple: The baseline PHX constructions for the PHPP U-Values Worksheet."""

    roof: PhxConstructionOpaque
    exposed_wall: PhxConstructionOpaque
    ground_wall: PhxConstructionOpaque
    exposed_floor: PhxConstructionOpaque
    ground_floor: PhxConstructionOpaque


class BaselineConstructionPHPPids(NamedTuple):
    """NamedTuple: Output of the add_baseline_constructions_to_phpp function."""

    roof: str
    exposed_wall: str
    ground_wall: str
    exposed_floor: str
    ground_floor: str


def create_baseline_constructions(
    baseline_code: BaselineCode,
    climate_zone: ClimateZones,
    use_group: Use_Groups,
    output: Callable = print,
) -> BaselinePHXConstructions:
    """Create the baseline constructions from the code baseline model."""

    # -- Roofs
    new_roof_phx_construction = PhxConstructionOpaque.from_total_u_value(
        baseline_code.get_baseline_roof_u_value(climate_zone, use_group),
        "BASELINE: ROOF",
    )

    # -- Exposed Walls (Above Grade)
    new_exposed_wall_phx_construction = PhxConstructionOpaque.from_total_u_value(
        baseline_code.get_baseline_exposed_wall_u_value(climate_zone, use_group),
        "BASELINE: EXPOSED WALL",
    )

    # -- Ground Walls (Below Grade)
    new_ground_wall_phx_construction = PhxConstructionOpaque.from_total_u_value(
        baseline_code.get_baseline_ground_wall_c_factor(climate_zone, use_group),
        "BASELINE: GROUND WALL",
    )

    # -- Exposed Floors
    new_exposed_floor_phx_construction = PhxConstructionOpaque.from_total_u_value(
        baseline_code.get_baseline_exposed_floor_u_value(climate_zone, use_group),
        "BASELINE: EXPOSED FLOOR",
    )

    # -- Ground Floors
    new_ground_floor_phx_construction = PhxConstructionOpaque.from_total_u_value(
        baseline_code.get_baseline_ground_floor_f_factor(climate_zone, use_group),
        "BASELINE: GROUND FLOOR",
    )

    return BaselinePHXConstructions(
        new_roof_phx_construction,
        new_exposed_wall_phx_construction,
        new_ground_wall_phx_construction,
        new_exposed_floor_phx_construction,
        new_ground_floor_phx_construction,
    )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create and add new baseline PHPP components."""

from PHX.PHPP import phpp_app
from PHX.model.constructions import PhxConstructionWindow
from PHX.PHPP.phpp_model.component_glazing import GlazingRow
from PHX.PHPP.phpp_model.component_frame import FrameRow


def add_new_baseline_window_glazing(
    phpp_conn: phpp_app.PHPPConnection, baseline_phx_window: PhxConstructionWindow
) -> str:
    """Add a new baseline PhxConstructionWindow Glazing to the PHPP Components Worksheet.

    Returns:
    --------
        str: The PHPP Glazing ID (ie: "01ud-MyGlass", etc..)
    """
    phpp_glazing_row = GlazingRow(phpp_conn.shape.COMPONENTS, baseline_phx_window)
    row_num = phpp_conn.components.first_empty_glazing_row_num
    phpp_glazing_id = phpp_conn.components.write_single_glazing(
        row_num, phpp_glazing_row
    )
    return phpp_glazing_id


def add_new_baseline_window_frame(
    phpp_conn: phpp_app.PHPPConnection, baseline_phx_window: PhxConstructionWindow
) -> str:
    """Add a new baseline PhxConstructionWindow Frame to the PHPP Components Worksheet.

    Returns:
    --------
        str: The PHPP Frame ID (ie: "01ud-MyFrame", etc..)
    """
    phpp_frame_row = FrameRow(phpp_conn.shape.COMPONENTS, baseline_phx_window)
    row_num = phpp_conn.components.first_empty_frame_row_num
    phpp_frame_id = phpp_conn.components.write_single_frame(row_num, phpp_frame_row)
    return phpp_frame_id



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to set baseline PHX Model lighting."""

from typing import Callable

from PHX.model.project import PhxProject

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.lighting_space_types import space_type_map


def set_baseline_lighting_installed_power_density(
    _phx_project: PhxProject, _baseline_code: BaselineCode, _output: Callable = print
) -> PhxProject:
    # TODO

    return _phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for setting the PHX Model envelope constructions to the baseline values. """

from typing import Callable

from PHX.model.enums.building import ComponentFaceType, ComponentExposureExterior
from PHX.model.project import PhxProject

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.options import ClimateZones, Use_Groups
from ph_baseliner.phx.constructions import create_baseline_constructions


def set_baseline_envelope_constructions(
    _phx_project: PhxProject,
    baseline_code: BaselineCode,
    climate_zone: ClimateZones,
    use_group: Use_Groups,
    output: Callable = print,
) -> PhxProject:
    """Set the PHX Model envelope constructions to the baseline values."""

    baseline_constructions = create_baseline_constructions(
        baseline_code, climate_zone, use_group
    )

    (
        roof,
        exposed_wall,
        ground_wall,
        exposed_floor,
        ground_floor,
    ) = baseline_constructions

    # -- Add all the new baseline constructions to the PHX project
    for construction in baseline_constructions:
        _phx_project.add_assembly_type(construction, _key=construction.identifier)

    # -- Set each envelope surface to use the right baseline construction
    for variant in _phx_project.variants:
        for opaque_component in variant.building.opaque_components:
            face_type = opaque_component.face_type
            face_exposure = opaque_component.exposure_exterior

            if face_type == ComponentFaceType.WALL:
                if face_exposure == ComponentExposureExterior.EXTERIOR:
                    opaque_component.set_assembly_type(exposed_wall)
                elif face_exposure == ComponentExposureExterior.SURFACE:
                    opaque_component.set_assembly_type(exposed_wall)
                else:
                    opaque_component.set_assembly_type(ground_wall)
            elif face_type == ComponentFaceType.ROOF_CEILING:
                opaque_component.set_assembly_type(roof)
            elif face_type == ComponentFaceType.FLOOR:
                if face_exposure == ComponentExposureExterior.EXTERIOR:
                    opaque_component.set_assembly_type(exposed_floor)
                elif face_exposure == ComponentExposureExterior.SURFACE:
                    opaque_component.set_assembly_type(exposed_floor)
                else:
                    opaque_component.set_assembly_type(ground_floor)

    return _phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Equivalence map between PHPP-space-type-names and Code-space-type-names."""

# -- TABLE C405.3.2(1) INTERIOR LIGHTING POWER ALLOWANCES: BUILDING AREA METHOD
# -- PHPP-Space-Type-Name : Code-Space-Type-Name
space_type_map = {
    "SINGLE OFFICE": "OFFICE",
    "GROUP OFFICE": "OFFICE",
    "OPEN-PLAN OFFICE": "OFFICE",
    "MEETING": "OFFICE",
    "COUNTER AREA": "RETAIL",
    "RETAIL": "RETAIL",
    "RETAIL/DEPARTMENT STORE WITH REFRIGERATED PRODUCTS": "RETAIL",
    "CLASSROOM": "SCHOOL/UNIVERSITY",
    "UNIVERSITY AUDITORIUM": "SCHOOL/UNIVERSITY",
    "BEDROOM": "DORMITORY",
    "HOTEL ROOM": "HOTEL/MOTEL",
    "CANTEEN": "DINING: CAFETERIA/FAST FOOD",
    "RESTAURANT": "DINING: FAMILY",
    "KITCHEN IN NON-RESIDENTIAL BUILDING": "DINING: FAMILY",
    "KITCHEN, STORAGE, PREPARATION": "DINING: FAMILY",
    "WC, SANITARY": "OFFICE",
    "OTHER HABITABLE ROOMS": "OFFICE",
    "SECONDARY AREAS": "OFFICE",
    "CIRCULATION AREA": "OFFICE",
    "STORAGE, SERVICES": "OFFICE",
    "SERVER ROOM": "OFFICE",
    "WORKSHOP (STANDING ACTIVITY)": "WORKSHOP",
    "WORKSHOP (PREDOMINANTLY STANDING ACTIVITY)": "WORKSHOP",
    "WORKSHOP (PREDOMINANTLY SEDENTARY ACTIVITY)": "WORKSHOP",
    "THEATRE AUDITORIUM": "PERFORMING ARTS THEATER",
    "THEATRE FOYER": "PERFORMING ARTS THEATER",
    "THEATRE STAGE": "PERFORMING ARTS THEATER",
    "FAIR, CONGRESS": "COURTHOUSE",
    "EXHIBITION": "MUSEUM",
    "LIBRARY READING ROOM": "LIBRARY",
    "OPEN ACCESS LIBRARY": "LIBRARY",
    "LIBRARY REPOSITORY": "LIBRARY",
    "SPORTS HALL": "SPORTS ARENA",
    "PARKING GARAGE": "PARKING GARAGE",
    "PUBLIC PARKING GARAGE": "PARKING GARAGE",
    "SAUNA AREA": "EXERCISE CENTER",
    "EXERCISE ROOM": "EXERCISE CENTER",
    "LABORATORY": "HEALTH CARE CLINIC",
    "EXAMINATION/TREATMENT ROOMS": "HEALTH CARE CLINIC",
    "SPECIAL CARE AREAS": "HEALTH CARE CLINIC",
    "CORRIDORS OF THE GENERAL CARE AREA": "HEALTH CARE CLINIC",
    "MEDICAL/THERAPEUTIC PRACTICES": "HEALTH CARE CLINIC",
    "WAREHOUSES": "WAREHOUSE",
}



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for creating and setting Baseline Constructions on the PHPP U-Values worksheet."""

from copy import copy

from PHX.PHPP import phpp_app

from ph_baseliner.phx.constructions import (
    BaselinePHXConstructions,
    BaselineConstructionPHPPids,
)


def add_baseline_constructions_to_phpp(
    phpp_conn: phpp_app.PHPPConnection, baseline_constructions: BaselinePHXConstructions
) -> BaselineConstructionPHPPids:
    """Add the baseline constructions to the PHPP U-Values Worksheet."""
    roof_phpp_id = phpp_conn.u_values.add_new_phx_construction(
        baseline_constructions.roof
    )
    exposed_wall_phpp_id = phpp_conn.u_values.add_new_phx_construction(
        baseline_constructions.exposed_wall
    )
    ground_wall_phpp_id = phpp_conn.u_values.add_new_phx_construction(
        baseline_constructions.ground_wall
    )
    exposed_floor_phpp_id = phpp_conn.u_values.add_new_phx_construction(
        baseline_constructions.exposed_floor
    )
    ground_floor_phpp_id = phpp_conn.u_values.add_new_phx_construction(
        baseline_constructions.ground_floor
    )

    return BaselineConstructionPHPPids(
        roof_phpp_id,
        exposed_wall_phpp_id,
        ground_wall_phpp_id,
        exposed_floor_phpp_id,
        ground_floor_phpp_id,
    )


def replace_u_values_with_baseline_constructions(
    phpp_conn: phpp_app.PHPPConnection, baseline_constructions: BaselinePHXConstructions
) -> None:
    """Replace the existing U-Values with the baseline constructions."""

    def is_roof(_input: str):
        """Return True if the R_si type is a 'Roof'."""
        return "ROOF" in str(_input).upper()

    def is_wall(_input: str):
        """Return True if the R_si type is a 'Wall'."""
        return "WALL" in str(_input).upper()

    def is_floor(_input: str):
        """Return True if the R_si type is a 'Floor'."""
        return "FLOOR" in str(_input).upper()

    def is_below_grade(_input: str):
        """Return True if the R_se type is a 'Ground'."""
        return "GROUND" in str(_input).upper()

    for row_num in phpp_conn.u_values.used_constructor_start_rows:
        constructor_name = phpp_conn.u_values.get_constructor_name(row_num)
        r_si_type = phpp_conn.u_values.get_constructor_r_si_type(row_num)
        r_se_type = phpp_conn.u_values.get_constructor_r_se_type(row_num)

        if is_roof(r_si_type):
            _phx_const = copy(baseline_constructions.roof)
        elif is_wall(r_si_type) and is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.ground_wall)
        elif is_wall(r_si_type) and not is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.exposed_wall)
        elif is_floor(r_si_type) and is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.ground_floor)
        elif is_floor(r_si_type) and not is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.exposed_floor)
        else:
            continue

        _phx_const.display_name = constructor_name

        # -- Update the PHPP
        phpp_conn.u_values.clear_single_constructor_data(row_num, False)
        phpp_conn.u_values.write_single_PHX_construction(_phx_const, row_num)
